/*
 * Assembly Language Debugger
 *
 * Copyright (C) 2000 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 * Additional changes to UI and addition of NASM based single line assembler
 * Copyright (C) 2005 Terry Loveall
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: command.c,v 1.23 2002/01/01 03:40:52 cosine Exp $
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <setjmp.h>
#include <signal.h>

#include "alddefs.h"
#include "command.h"
#include "defs.h"
#include "input.h"
#include "load.h"
#include "main.h"
#include "misc.h"
#include "msg.h"
#include "print.h"
#include "readln.h"
#include "signals.h"
#include "terminal.h"
#include "wgets.h"
#include "window.h"

/*
 * libOp includes
 */
#include "disasm.h"

/*
 * libString includes
 */
#include "alloc.h"
#include "Strn.h"

static void ProcessCommand(struct aldWorkspace *main_p, char *command);

static struct Command maincmds[] = {
  /*
   * Aliases - put these at the top since they are used
   * frequently
   */
  { "?", c_help, 0 },
  { "a", c_asm, C_FILELOADED },
  { "g", c_continue, C_ALIAS|C_FILELOADED|C_PTRACE },
  { "d", c_examine, C_ALIAS|C_FILELOADED },
  { "e", c_enter, C_FILELOADED },
  { "l", c_load, 0 },
  { "r", c_register, C_FILELOADED },
  { "s", c_next, C_ALIAS|C_FILELOADED|C_PTRACE },
  { "t", c_step, C_ALIAS|C_FILELOADED|C_PTRACE },
  { "u", c_disassemble, C_ALIAS|C_FILELOADED },
  { "w", c_writef, C_ALIAS|C_FILELOADED },

  { "bc", c_delete, 0 },
  { "bd", c_disable, 0 },
  { "be", c_enable, 0 },
  { "bp", c_break, C_FILELOADED },
  { "bi", c_ignore, 0 },
  { "bl", c_lbreak, 0 },
  { "bt", c_tbreak, C_FILELOADED },
  { "about", c_help, 0 },
  { "asm", c_asm, C_FILELOADED },
  { "go", c_continue, C_FILELOADED|C_PTRACE },
  { "dump", c_examine, C_ALIAS|C_FILELOADED },
  { "enter", c_enter, C_FILELOADED },
  { "file", c_file, C_FILELOADED },
  { "help", c_help, 0 },
  { "load", c_load, 0 },
  { "quit", c_quit, 0 },
  { "run", c_run, C_FILELOADED|C_PTRACE },
  { "set", c_set, 0 },
  { "unload", c_unload, C_FILELOADED },
  { "write", c_writef, C_FILELOADED },

  { 0, 0, 0 }
};

/*
initCommand()
  Initialize a command workspace

Return: pointer to workspace
*/

struct commandWorkspace *
initCommand()

{
  struct commandWorkspace *ws;

  ws = (struct commandWorkspace *) malloc(sizeof(struct commandWorkspace));
  if (!ws)
  {
    fprintf(stderr, "initCommand: malloc failed: %s\n",
      strerror(errno));
    return (0);
  }

  memset(ws, '\0', sizeof(struct commandWorkspace));

  ws->CmdPrompt = Strdup(ALD_PROMPT);

  return (ws);
} /* initCommand() */

/*
termCommand()
  Terminate a command workspace

Inputs: ws - workspace to terminate
*/

void
termCommand(struct commandWorkspace *ws)

{
  if (!ws)
    return;

  if (ws->CmdPrompt)
    free(ws->CmdPrompt);

  free(ws);
} /* termCommand() */

/*
procCommand()
  Enter infinite loop to wait for and process user commands

Inputs: main_p - main workspace
*/

void
procCommand(struct aldWorkspace *main_p)

{
  struct commandWorkspace *command_p = main_p->commandWorkspace_p;
  char buffer[MAXLINE + 1];
  char last[MAXLINE + 1]; /* last command entered */
  char *bufptr,
       *cmdptr;
  int GotLongJmp;

  *last = '\0';

  if (ModeConsole)
  {
    while (1)
    {
      GotLongJmp = setjmp(command_p->CmdParserEnv);
      if (GotLongJmp)
      {
        /*
         * In case we were in the middle of a print burst and the
         * user ^C'd, clear the burst flag
         */
        endPrintBurst(mainWorkspace_p->printWorkspace_p);

        restoreTerminal(&(main_p->terminalWorkspace_p->ParentAttributes));
        if (ModeConsole)
          printf("\n");
      }

      /*
       * Read a line from stdin
       */
      bufptr = ReadLine();
      assert(bufptr != 0);

      if ((*bufptr == '\0') || (*bufptr == '\n'))
      {
        /*
         * Use the last command entered - put 'last' back into
         * 'buffer' so it doesn't get destroyed by ProcessCommand()
         */
        strncpy(buffer, last, MAXLINE);
        cmdptr = buffer;
      }
      else
      {
        /*
         * Save this command for later use
         */
        strncpy(last, bufptr, MAXLINE);
        last[MAXLINE] = '\0';
        cmdptr = bufptr;
      }

      /*
       * Process given command
       */
      ProcessCommand(main_p, cmdptr);

      FreeLine(bufptr);
    } /* while (1) */
  }

#ifdef USE_CURSES

  if (ModeCurses)
  {
    assert(CurrentFrame != 0);

    while (1)
    {
      /*DisplayPrompt();*/

      RefreshFrames();

      wGets(CurrentFrame->window, buffer, MAXLINE);

      if (*buffer)
      {
        AddInput(buffer);
        ProcessCommand(main_p, buffer);
      }
    }
  } /* if (ModeCurses) */

#endif /* USE_CURSES */

} /* procCommand() */

/*
ProcessCommand()
  Process a given command

Inputs: main_p  - main workspace
        command - command

Return: none
*/

static void
ProcessCommand(struct aldWorkspace *main_p, char *command)

{
  int ac;
  char **av;
  struct Command *cptr;
  unsigned int flags;

  assert(command != 0);

  ac = SplitBuffer(command, &av);

  if (!ac)
    return;

  if (ac == (-1))
  {
    Print(P_COMMAND, "Unbalanced quotes");
    return;
  }

  flags = 0;
  cptr = FindCommand(maincmds, av[0], &flags);

  if (!cptr)
  {
    if (flags & C_AMBIGUOUS)
    {
      Print(P_COMMAND, "Ambiguous command: %s",
        av[0]);
    }
    else
    {
      Print(P_COMMAND, "Unknown command: %s (help shows a list)",
        av[0]);
    }

    MyFree(av);
    return;
  }

  if ((cptr->flags & C_FILELOADED) && !main_p->loadWorkspace_p->MapPtr)
  {
    Print(P_COMMAND, MSG_NOFILELOADED);
    MyFree(av);
    return;
  }

#ifdef USE_CURSES

  if (!ModeCurses && (cptr->flags & C_CURSES))
  {
    Print(P_COMMAND,
      "The command \"%s\" works only in curses mode",
      cptr->cmd);
    MyFree(av);
    return;
  }

#endif /* USE_CURSES */

  if (cptr->flags & C_PTRACE)
  {
    /*
     * The command we are about to execute will be using the ptrace()
     * system call - do not catch SIGINT, in case the console user
     * hits ^C while the ptraced process is running - we don't
     * want to longjmp() back to procCommand() in the middle
     * of the ptrace()
     */
    signal(SIGINT, SIG_IGN);
  }
  else
  {
    /*
     * Hook SIGINT to our signal handler, in case we are about
     * to execute a command which bursts a lot of info - the user
     * will have the ability to hit CONTROL-C and stop the execution
     * of the command.
     * This must be set up after the setjmp() call in procCommand()
     * or things will be very bad if ^C is pressed beforehand
     * (ie: the SigHandler longjmp() call will jump to who knows where)
     */
    signal(SIGINT, SigHandler);
  }

  /*
   * Execute the command
   */
  (*cptr->funcptr)(ac, av);

  MyFree(av);
} /* ProcessCommand() */

/*
FindCommand()
 Attempt to find the command "name" in the array "cmdlist"

Inputs: cmdlist - array to look in
        name    - command name
        flags   - bitmasks

Return: pointer to the index containing "name" if found,
        otherwise NULL.
        If the command is found, but there is more than 1 match
        (ambiguous), set the C_AMBIGUOUS bit in 'flags'
*/

struct Command *
FindCommand(struct Command *cmdlist, char *name, unsigned int *flags)

{
  struct Command *cmdptr, *tmp;
  int matches, /* number of matches we've had so far */
      clength;

  assert(name != 0);
  assert(cmdlist != 0);

  tmp = NULL;
  matches = 0;
  clength = strlen(name);
  for (cmdptr = cmdlist; cmdptr->cmd; ++cmdptr)
  {
    if (!Strncasecmp(name, cmdptr->cmd, clength))
    {
      if (clength == strlen(cmdptr->cmd))
      {
        /*
         * name and cmdptr->cmd are the same length, so it
         * must be an exact match, don't search any further
         */
        matches = 0;
        break;
      }
      tmp = cmdptr;
      ++matches;
    }
  }

  /*
   * If matches > 1, name is an ambiguous command, so the
   * user needs to be more specific
   */
  if ((matches == 1) && (tmp))
    cmdptr = tmp;

  if (cmdptr->cmd)
    return (cmdptr);

  if (matches != 0)
    *flags |= C_AMBIGUOUS;

  return (0);
} /* FindCommand() */
